import dataclasses
import json
import os


@dataclasses.dataclass
class Settings:
    host: str
    port: int
    logLevel: str


settings: Settings | None = None


def get_settings():
    """
    Liest die Einstellungen aus der settings.json und gibt diese zurück
    :return: Einstellungen
    """

    global settings

    if settings is None:
        settings_file = "settings.json"
        if not os.path.exists(settings_file) or not os.path.isfile(settings_file):
            settings = Settings(host="0.0.0.0", port=8075, logLevel="INFO")

        if settings is None:
            with open(settings_file, "r") as file:
                json_data = json.load(file)

            settings = Settings(**json_data)

    return settings



from pydantic import BaseModel, Field
from typing import Callable, ForwardRef
import polars as pl
from .enums import FilterOperator, DataType

type LoadOptionsInt = ForwardRef("LoadOptions", "odm_models_query")
type LoadPropertyInt = ForwardRef("LoadProperty", "odm_models_query")


class DefinitionProperty(BaseModel):
    alias: str = Field(exlcude=True)
    name: str = Field(serialization_alias="name")

    type: DataType = Field(serialization_alias="type")

    key: str | None = Field(None, serialization_alias="key")
    description: str | None = Field(None, serialization_alias="description")
    format_string: str | None = Field(None, serialization_alias="formatString")
    regex: str | None = Field(None, serialization_alias="regex")

    can_aggregate: bool | None = Field(None, serialization_alias="canAggregate")

    """
    Diese Eigenschaften können nicht direkt gefilter werden, da diese erst im Nachhinein berechnet werden.
    Falls doch ein Filter kommt, werden diese mit True bestätigt. Filterung muss, wenn notwendig  mit Hilfe
    von Variablen erfolgen.
    # TODO Filter (siehe Beschreibung)
    """
    agg_expr: Callable[[LoadOptionsInt, LoadPropertyInt, bool], pl.Expr] | None = Field(None, exclude=True)
    final_expr: Callable[[LoadOptionsInt, LoadPropertyInt], pl.Expr] | None = Field(None, exclude=True)


class DefinitionVariable(BaseModel):
    alias: str = Field(exclude=True)
    name: str = Field(serialization_alias="name")

    type: DataType = Field(serialization_alias="type")

    supported_filter_operators: FilterOperator = Field(serialization_alias="supportedFilterOperators")

    key: str | None = Field(None, serialization_alias="key")
    description: str | None = Field(None, serialization_alias="description")
    format_string: str | None = Field(None, serialization_alias="formatString")
    regex: str | None = Field(None, serialization_alias="regex")

    is_required: bool | None = Field(None, serialization_alias="isRequired")


class DefinitionFilterBlock(BaseModel):
    alias: str = Field(exclude=True)
    name: str = Field(alias="name")

    key: str | None = Field(None, serialization_alias="key")
    description: str | None = Field(None, serialization_alias="description")


class Definition(BaseModel):
    key: str = Field(serialization_alias="key")
    name: str = Field(serialization_alias="name")
    description: str | None = Field(None, serialization_alias="description")

    properties: list[DefinitionProperty] = Field(serialization_alias="properties")

    variables: list[DefinitionVariable] | None = Field(None, serialization_alias="variables")
    filter_blocks: list[DefinitionFilterBlock] | None = Field(None, serialization_alias="filterBlocks")


class DefinitionCompact(BaseModel):
    key: str = Field(serialization_alias="key")
    name: str = Field(serialization_alias="name")
    description: str | None = Field(None, serialization_alias="description")



import importlib
import os
import polars as pl
import pathlib
import yaml
from .query import add_query
from .query_utils import filter_aggregate_convert
from .enums import DataType
from .models_definition import Definition, DefinitionProperty
from .models_query import LoadOptions
from .caching import data_cache, get_cached_data


def auto_import_path(path: str):
    __auto_import_py(path)
    __auto_import_yaml(path)


def __auto_import_py(path: str):
    p = pathlib.Path.cwd() / path
    files = p.glob("*.py")

    for file in files:
        r = str(file.relative_to(pathlib.Path.cwd())).replace("/", ".").replace("\\", ".")[:-3]
        module = importlib.import_module(str(r))

        if "add" in module.__dict__ and callable(module.__dict__["add"]):
            module.add()


def __auto_import_yaml(path: str):
    p = pathlib.Path.cwd() / path
    files = list(p.glob("*.yml")) + list(p.glob("*.yaml"))

    for file in files:
        add_yaml(str(file.relative_to(pathlib.Path.cwd())))


def add_yaml(file: str):
    file = pathlib.Path.cwd() / file

    with open(file, "r", encoding="utf-8") as f:
        dict_def: dict[str, any] = yaml.safe_load(f)

    dict_properties: dict[str, any] = dict_def["properties"]
    dict_cache: dict[str, any] | None = dict_def.get("cache", None)

    conn: str | None = dict_def["conn"]
    sql: str | None = dict_def["sql"]

    properties: list[DefinitionProperty] = []
    for key, value in dict_properties.items():
        t = value.get("type", "string")

        properties.append(DefinitionProperty(
            alias=key,
            name=value.get("name", key),
            type=__convert_str_to_data_type(t),
            description=value.get("description", None),
            format_string=value.get("formatString", None),
            regex=value.get("regex", None),
            can_aggregate=value.get("canAggregate", False)
        ))

    d = Definition(
        key=dict_def["key"],
        name=dict_def["name"],
        description=dict_def.get("description", None),

        properties=properties
    )

    uri = os.getenv(conn)
    cache_key: str | None = dict_cache.get("key", None) if dict_cache is not None else None

    def get_data():
        return pl.read_database(sql, uri)

    async def f(options: LoadOptions):
        df = get_data() if cache_key is None else await get_cached_data(cache_key)
        return filter_aggregate_convert(options, df)

    if cache_key is not None:
        cache_name: str | None = dict_cache.get("name", cache_key)
        cache_timer = dict_cache.get("refreshTime", ["02:00"])
        cache_recreate = dict_cache.get("recreate", True)
        cache_lazy = dict_cache.get("lazy", True)
        depend_of = dict_cache.get("dependOf", None)

        if isinstance(depend_of, str):
            depend_of = [depend_of]

        if isinstance(cache_timer, str):
            cache_timer = [cache_timer]

        # Cache für die Definition aktivieren
        data_cache(cache_key, cache_name, cache_timer, cache_recreate, cache_lazy, depend_of)(get_data)

    add_query(d, f)


def __convert_str_to_data_type(string: str) -> DataType:
    match string:
        case "string":
            return DataType.string
        case "int":
            return DataType.int
        case "decimal":
            return DataType.decimal
        case "datetime":
            return DataType.datetime
        case "bytes":
            return DataType.bytes



import polars as pl
from .models_definition import Definition
from .enums import PropertyAggregateType
from .filter_polars_dataframe_visitor import FilterPolarsDataframeVisitor
from .models_query import LoadOptions, LoadProperty, FilterGroup
from .models_definition import DefinitionProperty


def get_date_columns(d: Definition) -> list[str]:
    """
    Liefert alle DateTime-Spalten (dessen Alias) einer Definition
    :param d: Definition 
    :return: list[str]
    """

    return [col.alias for col in d.properties if col.type == "datetime"]


def convert_to_dataframe(df: pl.DataFrame | pl.LazyFrame) -> pl.DataFrame:
    """
    Falls das übergebene DataFrame ein LazyFrame ist, dann wird dies in ein "normales" DataFrame konvertiert.
    Falls es ein "normales" DataFrame ist, dann wird dieses 1zu1 zurückgegeben.
    :param df: DataFrame oder LazyFrame
    :return: DataFrame
    """

    if isinstance(df, pl.LazyFrame):
        return df.collect()

    return df


def filter_aggregate_convert(options: LoadOptions, df: pl.DataFrame, agg_expressions: list[pl.Expr] | None = None) -> str:
    """
    Kombiniert die Funktionen filter_dataframe, aggregate_dataframe und convert_dataframe_to_result
    :param options: LoadOptions
    :param df: DataFrame
    :param agg_expressions: Angabe von zusätzlichen Aggregate-Expressions. .alias(xxx) muss angegeben werden.
    :return: DataFrame
    """

    df = filter_dataframe(options, df)
    df = aggregate_dataframe(options, df, agg_expressions=agg_expressions)
    df = final_expr_dataframe(options, df)
    return convert_dataframe_to_result(options, df)


def filter_dataframe(options: LoadOptions, df: pl.DataFrame) -> pl.DataFrame:
    """
    Filter das DataFrame auf Basis der Filter in LoadOptions
    :param options: LoadOptions
    :param df: DataFrame
    :return: DataFrame
    """

    visitor = FilterPolarsDataframeVisitor(options.definition, df)

    if options.filter:
        r = options.filter.visit(visitor)
        df = df.filter(r)

    if options.relation_filter:
        # Diese Filter stelle eine Besonderheit dar, da diese fixiert sind in:
        # Wenn es ein FilterList ist, dann wird nur nach einem Element gefilter mit mehreren Werten 
        # => somit kein Problem

        # Wenn es ein FilterGroup ist, dann besteht die Beziehung aus mehreren Eigenschaften
        # => dies könnte theoretisch über die normale Group.Or Logik laufen, allerdings stürzt Polars ab, wenn
        #    es zu viele Filter sind.
        # => daher Logik über struct. Dann kann wieder mit is_in gefiltert werden.

        # Wenn der Relation-Filter allerdings nur für eine Zeile ist, dann kommt keine Gruppe daher
        # => dann kann der Filter direkt angewendet werden

        f = options.relation_filter

        if isinstance(f, FilterGroup):
            group: FilterGroup = f

            if len(group.filters) == 1:
                f = group.filters[0]

        if isinstance(f, FilterGroup) and len(f.filters) > 0 and isinstance(f.filters[0], FilterGroup):
            group: FilterGroup = f

            prop_dict = {p.key: p.alias for p in options.definition.properties}
            rel = [{prop_dict[i.key]: i.value for i in f.filters} for f in group.filters]
            df = df.filter(pl.struct(rel[0].keys()).is_in(rel))

        else:
            df = df.filter(f.visit(visitor))

    return df


def aggregate_dataframe(options: LoadOptions, df: pl.DataFrame, agg_expressions: list[pl.Expr] | None = None) -> pl.DataFrame:
    """
    Aggregiert das DataFrame auf Basis der Properties in LoadOptions.
    Wenn min. ein LoadProperty existiert, dass eine einen aggregate_type != PropertyAggregateType.no hat
    oder ein LoadProperty existiert, dass eine agg_expr hat
    oder wenn der agg_expressions-Parameter Werte enthält,
    wird eine Gruppierung und Aggregierung vorgenommen.
    :param options: LoadOptions
    :param df: DataFrame
    :param agg_expressions: Angabe von zusätzlichen Aggregate-Expressions. .alias(xxx) muss angegeben werden.
    :return: DataFrame
    """

    if agg_expressions is None:
        agg_expressions = []

    aggregates: list[tuple[DefinitionProperty, LoadProperty]] = []
    groups: list[tuple[DefinitionProperty, LoadProperty]] = []

    for p in options.properties:
        prop = options.get_definition_property(p.key)

        if prop.agg_expr is None and prop.final_expr is not None:
            continue

        if p.aggregate_type == PropertyAggregateType.no and prop.agg_expr is None:
            groups.append((prop, p))
        else:
            aggregates.append((prop, p))

    aggregate_count = len(aggregates) + len(agg_expressions)

    if aggregate_count == 0:
        agg_list = []

        for prop, load_prop in groups:
            if prop.agg_expr is not None:
                agg_list.append(prop.agg_expr(options, load_prop, False).alias(load_prop.field_name))

        agg_list += agg_expressions

        if len(agg_list) > 0:
            df = df.with_columns(agg_list)

        return df

    if len(groups) == 0:
        agg_list = []

        for prop, load_prop in aggregates:

            if prop.agg_expr is not None:
                agg_list.append(prop.agg_expr(options, load_prop, True).alias(load_prop.field_name))
            else:
                col = convert_aggregate_type(
                    pl.col(prop.alias).alias(load_prop.field_name),
                    load_prop.aggregate_type)

                agg_list.append(col)

        df = df.select(agg_list)
    else:
        agg_list = []
        group_list = []

        for prop, load_prop in aggregates:
            if prop.agg_expr is not None:
                agg_list.append(prop.agg_expr(options, load_prop, True).alias(load_prop.field_name))
            else:
                agg_list.append(
                    convert_aggregate_type(
                        pl.col(prop.alias),
                        load_prop.aggregate_type
                    )
                    .alias(load_prop.field_name)
                )

        agg_list += agg_expressions

        for prop, load_prop in groups:
            group_list.append(pl.col(prop.alias).alias(load_prop.field_name))

        df = df.group_by(group_list).agg(agg_list)

    return df


def final_expr_dataframe(options: LoadOptions, df: pl.DataFrame) -> pl.DataFrame:
    """
    Führt die Final-Expressions lt. Definition aus
    :param options: LoadOptions
    :param df: DataFrame
    :return: DataFrame
    """

    for p in options.properties:
        prop = options.get_definition_property(p.key)

        if prop.final_expr is None:
            continue

        df = df.with_columns(prop.final_expr(options, p).alias(p.field_name))

    return df


def convert_dataframe_to_result(options: LoadOptions, df: pl.DataFrame) -> str:
    """
    Konvertiert das DataFrame in einen JSON-String, der vom ODM verarbeitet wird
    :param options: LoadOptions
    :param df: DataFrame
    :return: str
    """

    if options.max_rows:
        df = convert_to_dataframe(df).head(options.max_rows)

    cols = []
    for p in options.properties:
        if p.field_name in df.columns:
            cols.append(p.field_name)
            continue

        if p.alias in df.columns:
            cols.append(pl.col(p.alias).alias(p.field_name))
            continue

        raise Exception(f"Property {p.key} not found in dataframe")

    df = df.select(cols)

    return convert_to_dataframe(df).write_json()


def convert_aggregate_type(col: pl.Expr, aggregate_type: PropertyAggregateType) -> pl.Expr:
    """
    Gibt die Polars-Aggregate-Funktion für den ODM PropertyAggregateType zurück
    :param col: expr
    :param aggregate_type: PropertyAggregateType 
    :return: expr
    """

    match aggregate_type:
        case PropertyAggregateType.no:
            return col.first()
        case PropertyAggregateType.sum:
            return col.sum()
        case PropertyAggregateType.count:
            return col.count()
        case PropertyAggregateType.avg:
            return col.mean()
        case PropertyAggregateType.min:
            return col.min()
        case PropertyAggregateType.max:
            return col.max()
        case PropertyAggregateType.stdev:
            return col.std()



from enum import Flag, Enum


class FilterOperator(Flag):
    no = 0
    equals = 1
    not_equals = 2
    greater_than = 4
    greater_than_or_equals = 8
    less_than = 16
    less_than_or_equals = 32
    value_in = 64
    value_not_in = 128
    like = 256
    not_like = 512
    between = 1024
    not_between = 2048
    group_and = 4096
    group_or = 8192
    binary_filter = (equals
                     | not_equals
                     | greater_than
                     | greater_than_or_equals
                     | less_than
                     | less_than_or_equals
                     | like
                     | not_like)
    between_filter = between | not_between
    list_filter = value_in | value_not_in
    group_filter = group_and | group_or


class FilterBinaryOperator(int, Enum):
    equals = 0,
    not_equals = 1
    like = 2
    not_like = 3
    greater_than = 4
    greater_than_or_equals = 5
    less_than = 6
    less_than_or_equals = 7


class FilterBetweenOperator(int, Enum):
    between = 0
    not_between = 1


class FilterListOperator(int, Enum):
    value_in = 0
    value_not_in = 1


class FilterGroupOperator(int, Enum):
    group_and = 0
    group_or = 1


class DataType(str, Enum):
    string = "string"
    int = "int"
    decimal = "decimal"
    datetime = "datetime"
    bytes = "bytes"


class PropertyAggregateType(int, Enum):
    no = 0
    sum = 1
    max = 2
    min = 3
    avg = 4
    count = 5
    stdev = 6



from typing import Callable
from .filter_datatype_visitor import FilterDatatypeVisitor
from .models_definition import Definition, DefinitionCompact
from .models_query import LoadOptions
from .logger import log_error, log_exception
from .utils import create_uuid_from_string, timer

type DefinitionCallback = Callable[[], Definition] | Definition
type LoadCallback = Callable[[LoadOptions], any]

queries: list[tuple[DefinitionCallback, LoadCallback]] = []


def add_query(d: DefinitionCallback, f: LoadCallback):
    """
    Fügt ein Query in das Query-Repository ein
    :param d: Definition oder Funktion, die eine Definition zurückgibt (falls diese dynamisch ist)
    :param f: Funktion die aufgerufen wird, wenn die Daten ermittelt werden
    :return: None
    """

    if d is None:
        msg = "add_query: Definition darf nicht None sein"
        log_error(msg)
        raise Exception(msg)
    if f is None:
        msg = "add_query: Funktion darf nicht None sein"
        log_error(msg)
        raise Exception(msg)

    if isinstance(d, Definition):
        d = __validate_definition(d)

    queries.append((d, f))


def get_queries():
    """
    Gibt alle Queries (in kompakter Form) zurück
    :return: DefinitionCompact[]
    """

    d = map(lambda q: __get_definition(q[0]), queries)

    return map(lambda d: DefinitionCompact(
        key=d.key,
        name=d.name,
        description=d.description
    ), d)


def get_query(key):
    """
    Gibt eine Query zurück
    :param key: Key der Definition
    :return: Definition | None
    """

    d = next(filter(lambda d: d.key == key, map(lambda q: q[0], queries)), None)
    if d is None:
        log_error(f"get_query: Definition mit Key {key} nicht gefunden")
        return None

    d = __get_definition(d)
    return d


async def get_query_result(options: LoadOptions) -> str:
    """
    Ermittelt die Daten für eine Query
    :param options: LoadOptions
    :return: JSON-String mit den Ergebnissen
    """

    q = next(filter(lambda x: x[0].key == options.key, queries))
    if not q:
        log_error(f"[{options.identity_key}]  get_query_result: Query mit Key {options.key} nicht gefunden")
        return "[]"

    options.definition = __get_definition(q[0])
    __validate_load_options(options)

    try:
        text = f"[{options.identity_key}]  Query {options.definition.name}"
        return await timer(text, lambda: q[1](options))
    except Exception:
        log_exception(f"[{options.identity_key}]  Fehler bei Query {options.definition.name}")
        return "[]"


def __get_definition(d: DefinitionCallback) -> Definition:
    if isinstance(d, Definition):
        return d
    else:
        return __validate_definition(d())


def __validate_definition(d: Definition):
    for p in d.properties:
        if p.key is None:
            p.key = create_uuid_from_string(f"{d.key}_{p.alias}")

    if d.variables:
        for v in d.variables:
            if v.key is None:
                v.key = create_uuid_from_string(f"{d.key}_var_{v.alias}")

    if d.filter_blocks:
        for fb in d.filter_blocks:
            if fb.key is None:
                fb.key = create_uuid_from_string(f"{d.key}_fb_{fb.alias}")

    return d


def __validate_load_options(options: LoadOptions):
    d: Definition = options.definition

    alias_dict = {p.key: p.alias for p in d.properties}

    for p in options.properties:
        p.alias = alias_dict[p.key]

    visitor = FilterDatatypeVisitor(options)

    if options.filter:
        options.filter.visit(visitor)

    if options.relation_filter:
        options.relation_filter.visit(visitor)

    for v in options.variables:
        v.visit(visitor)



from pydantic import BaseModel, Field
import datetime


class Cache(BaseModel):
    key: str = Field(serialization_alias="key")
    name: str = Field(serialization_alias="name")
    last_refresh: datetime.datetime | None = Field(None, serialization_alias="lastRefresh")
    lazy: bool = Field(serialization_alias="lazy")
    is_invalidated: bool = Field(serialization_alias="isInvalidated")
    size_mb: float | None = Field(None, serialization_alias="sizeMB")
